import json


class Response(object):
    """
    http response content
    """

    def __init__(self, http_code, actual_response):
        self.http_code = http_code
        self.actual_response = actual_response

    def to_string(self):
        return '{"http_code":' + str(self.http_code) + ',"actual_response":' + json.dumps(
            self.actual_response) + '}'



from pdr_python_sdk.errors import IllegalArgument


class RepoConfigBuilder(object):
    def __init__(self, repo_name):
        self.config = {}
        self.repo_name = repo_name

    def set_desc(self, desc):
        self.config["description"] = desc
        return self

    def set_retention(self, retention):
        self.config["retention"] = int(retention)
        return self

    def enable_lifecycle(self):
        self.config["lifecyclePolicyEnable"] = True
        return self

    def disable_lifecycle(self):
        self.config["lifecyclePolicyEnable"] = False
        return self

    def set_repo_replicas(self, num):
        self.config["repoReplicas"] = num
        return self

    def set_write_replicas(self, num):
        self.config["writeReplicas"] = num
        return self

    def set_write_shards(self, num):
        self.config["writeShards"] = num
        return self

    def set_write_refresh_interval(self, seconds):
        self.config["writeRefreshIntervalInSeconds"] = seconds
        return self

    def init_rollover(self):
        self.config["rollover"] = {}
        return self

    def set_rollover(self, rollover):
        self.config["rollover"] = rollover
        return self

    def init_lifecycle(self):
        self.config["lifecyclePolicy"] = {
            "repoName": self.repo_name,
            "phases": {
                "hot": {
                    "phaseName": "hot"
                },
                "warm": {
                    "phaseName": "warm"
                },
                "cold": {
                    "phaseName": "cold"
                }
            }
        }
        return self

    def set_lifecycle_ages(self, lifecycle):
        self.config["lifecyclePolicy"]["phases"]["hot"]["maxAge"] = lifecycle.get(
            "hot")
        self.config["lifecyclePolicy"]["phases"]["warm"]["maxAge"] = lifecycle.get(
            "warm")
        self.config["lifecyclePolicy"]["phases"]["cold"]["maxAge"] = lifecycle.get(
            "cold")
        return self

    def build(self):
        ok, reason = self.check()
        if not ok:
            raise IllegalArgument(reason)
        return self.config

    """"
    TODO validate the config
    """

    def check(self):
        return check_repo_config(self.config)


def check_repo_config(repo_config):
    if repo_config.get("lifecyclePolicyEnable"):
            if not repo_config.get("lifecyclePolicy"):
                return (False, "lifecyclePolicy should be provided when lifecyclePolicyEnable is True")
            if not repo_config["lifecyclePolicy"]["phases"]["hot"].get("maxAge"):
                return (False, "hot/wam/cold maxAge should be provided when lifecyclePolicyEnable is True")
            if not repo_config["lifecyclePolicy"]["phases"]["warm"].get("maxAge"):
                return (False, "hot/wam/cold maxAge should be provided when lifecyclePolicyEnable is True")
            if not repo_config["lifecyclePolicy"]["phases"]["cold"].get("maxAge"):
                return (False, "hot/wam/cold maxAge should be provided when lifecyclePolicyEnable is True")

    if not repo_config.get("lifecyclePolicyEnable"):
        if "retention" not in repo_config or not repo_config.get("retention"):
            return (False, "retention should be provided when lifecyclePolicyEnable is False")
        if repo_config.get("lifecyclePolicy"):
            return (False, "lifecyclePolicy should not be provided when lifecyclePolicyEnable is False")
    return (True, "")



import logging
import sys
import os

from .common.logging import config_logging


class OnDemandAction(object):
    """
    One time action which process starts on new request and ends after handling it.
    """

    def on_request(self, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        """
        method used to handle request.
        """
        raise NotImplemented('method [on_request()] is not yet implemented')


def run(clz, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):

    # config logging
    config_logging()

    if argv is None:
        argv = sys.argv

    # add $APP_ROOT/bins/libs to path
    app_root_dir = argv[1]
    lib_name = os.path.sep.join([app_root_dir, 'bins', 'libs'])
    sys.path.insert(0, lib_name)

    logging.info('app root dir is: {}'.format(app_root_dir))
    logging.info('running script using class: [' + str(clz) + ']')

    try:
        clz().on_request(argv, input_stream, output_stream)
    except Exception as err:
        logging.exception(err)



import os
import logging


def config_logging(level=None, filename=None):
    if filename is None:
        log_dir = '../log'
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        filename = '../log/phoenix_app.log'
        if not os.path.exists(filename):
            open(filename, 'w').close()

    if level is None:
        level = 'INFO'

    logging.basicConfig(filename=filename,
                        level=level,
                        format="%(asctime)s %(name)s %(levelname)s %(message)s",
                        datefmt='%Y-%m-%d  %H:%M:%S %a')



from pdr_python_sdk.storage import *

if __name__ == "__main__":
    c = connect()
    storageTable = c.storage("cluster_monitoring")
    
    storageTable.create(data='{"name":"sdk_new7", "schema":"des TEXT"}')
    response = storageTable.get_tables()
    print("get_tables response=", str(response))
    
    response = storageTable.delete_table("sdk_new7")
    #response = storageTable.get_tables()
    print("delete_table response=", str(response)) 
    
    storageTable.delete_table("sdk_new8")
    storageTable.create(data='{"name":"sdk_new8", "schema":"title TEXT, des TEXT"}')
    response = storageTable.get_tables()
    
    storageTableData = storageTable.data("sdk_new8")
    response = storageTableData.insert({"fields": "title,des", "values": "\"lihong\",\"sdk_test\""})
    #response = storageTableData.query()
    print("insert sdk_new8 response=", str(response))  
    
    response = storageTableData.query(query="title=\"lihong\"")
    print("query sdk_new8 response=", str(response))
    
    response = storageTableData.updateByQuery('{"query": "title=\\"lihong\\"", "value": "des=\\"sdk_test_new\\""}')
    #response = storageTableData.query(query="title=\"lihong\"")
    print("updateByQuery sdk_new8 response=", str(response))
    
    response = storageTableData.delete(query="title=\"lihong\"")
    #response = storageTableData.query(query="title=\"lihong\"")
    print("delete query sdk_new8 response=", str(response))    
    
    storageTableData = storageTable.data("lihongtest")
    
    response = storageTableData.delete()
    #response = storageTableData.query()
    print("delete response=", str(response))      
    
    response = storageTableData.insert({"fields": "des", "values": "\"sdk_test\""})
    #response = storageTableData.query()
    print("insert response=", str(response))
    
    response = storageTableData.update(1, '{"value": "des=\\"sdk_new\\""}')
    #response = storageTableData.query_by_id(1)
    print("update response=", str(response)) 
    
    #storageTableData.delete_by_id(1)
    #response = storageTableData.query_by_id(1)
    #print("delete_by_id response=", str(response)) 
    
    response = storageTableData.delete(query="id=1")
    print("delete by query response=", str(response))    
    
    response = storageTableData.delete()
    #response = storageTableData.query()
    print("delete response=", str(response))     
    
    response = storageTableData.batch_save(records={"fields": "des", "values": ["\"sdk_test1\"","\"sdk_test2\"","\"sdk_test3\""]})
    #response = storageTableData.query(pageNo=1, pageSize=2)
    print("response=", str(response))   
    response = storageTableData.query(query="des=\"sdk_test1\"")
    print("query response=", str(response))    
    response = storageTableData.query()
    print("batch_save response=", str(response))   


#!/usr/bin/env python

import sys

from ...spl import SplStreamingBatchCommand
from ...on_demand_action import run

class Foobar(SplStreamingBatchCommand):
    def streaming_handle(self, lines):
        for line in lines:
            line['foo'] = 'bar'
        return lines


if __name__ == '__main__':
    run(Foobar, sys.argv, sys.stdin.buffer, sys.stdout.buffer)


import urllib3
import json
import time
from urllib.parse import urlencode
from pdr_python_sdk.errors import *
from pdr_python_sdk.entity import RepoConfigBuilder

# If you change these, update the docstring
# on _uri as well.
DEFAULT_HOST = "localhost"
DEFAULT_PORT = "8000"
DEFAULT_SCHEME = "http"
API_ROOT_PREFIX = "/api/v1"

# Request Header
DEFAULT_CONTENT_TYPE_KEY = 'Content-Type'
DEFAULT_CONTENT_TYPE_VALUE = 'application/json'
TOKEN_AUTH_HEADER_KEY = "Authorization"
TOKEN_SESSION_HEADER_KEY = "X-App-Session-Key"
DEFAULT_ENCODING = "utf-8"

# Request Path

# Data related
PATH_DATA_UPLOAD = "/data"

# SPL QUERY related
PATH_SPL_QUERY_JOB = "/jobs"
PATH_SPL_QUERY_JOB_STATUS = "/jobs/{}"
PATH_SPL_QUERY_JOB_EVENTS = "/jobs/{}/events"
PATH_SPL_QUERY_JOB_TIMELINE = "/jobs/{}/timeline"
PATH_SPL_QUERY_JOB_SUMMARY = "/jobs/{}/summary"
PATH_SPL_QUERY_JOB_RESULTS = "/jobs/{}/results"

# Repo related
PATH_REPOS = "/repos"
PATH_SINGLE_REPO = "/repos/{}"


def connect(**kwargs):
    return PandoraConnection(**kwargs)


class PandoraConnection(object):
    """
    Encapsulate connection pool, to make request to pandora server
    """

    def __init__(self, **kwargs):
        """
        Constructor of PandoraConnection, return a connection instance

        :param scheme: http or https
        :type scheme: ``string``
        :param host: the pandora server host
        :type host: ``string``
        :param port: the pandora server port
        :type port: ``integer``
        :param root_path: the pandora api root path, default is /api/v1
        :type root_path: ``string``
        :param token: required if using token to access
        :type token: ``string``
        :param session_key: required if using session_key to access (mostly in app)
        :type session_key: ``string``
        :param headers: additional headers to access
        :type headers: ``map``
        """
        self.scheme = kwargs.get("scheme", DEFAULT_SCHEME)
        self.host = kwargs.get("host", DEFAULT_HOST)
        self.port = int(kwargs.get("port", DEFAULT_PORT))
        self.root_path = kwargs.get("root_path", API_ROOT_PREFIX)
        self.root_url = urllib3.util.Url(
            host=self.host, scheme=self.scheme, port=self.port, path=self.root_path)

        self.token = kwargs.get("token", "")
        self.session_key = kwargs.get("session_key", "")

        self.additional_headers = kwargs.get("headers", {})
        if DEFAULT_CONTENT_TYPE_KEY not in self.additional_headers:
            self.additional_headers[DEFAULT_CONTENT_TYPE_KEY] = DEFAULT_CONTENT_TYPE_VALUE
        if len(self.token) > 0 and TOKEN_AUTH_HEADER_KEY not in self.additional_headers:
            self.additional_headers[TOKEN_AUTH_HEADER_KEY] = self.token
        if len(self.session_key) > 0 and TOKEN_SESSION_HEADER_KEY not in self.additional_headers:
            self.additional_headers[TOKEN_SESSION_HEADER_KEY] = self.session_key
        self.http = urllib3.PoolManager(headers=self.additional_headers)

    def request(self, method, subpath, fields=None, headers=None, **urlopen_kw):
        """
        Base request method, encapsulate `PoolManager.request` basic request.
        """
        resp = self.http.request(
            method, self.root_url.url + subpath, fields=fields, headers=headers, **urlopen_kw)
        if resp.status >= 500:
            raise ServerError("""Server Error Status: {}, Error Reason: {}, Body: {}""".format(
                resp.status, resp.reason, resp.data.decode(DEFAULT_ENCODING)))

        if resp.status == 404:
            raise NotFound("""Not Found Error Status: {}, Error Reason: {}, Body: {}""".format(
                resp.status, resp.reason, resp.data.decode(DEFAULT_ENCODING)))

        if resp.status >= 400:
            raise BadRequest("""Bad Request Error Status: {}, Error Reason: {}, Body: {}""".format(
                resp.status, resp.reason, resp.data.decode(DEFAULT_ENCODING)))

        return resp

    def request_json(self, method, subpath, fields=None, headers=None, **urlopen_kw):
        resp = self.request(method, subpath, fields=fields,
                            headers=headers, **urlopen_kw)
        return decode_json(resp.data)

    def post(self, subpath, body, fields=None):
        encoded_data = encode_json(body)
        if fields is not None:
            return self.request_json('POST', subpath + '?' + urlencode(fields), body=encoded_data)
        else:
            return self.request_json('POST', subpath, body=encoded_data)

    def get(self, subpath, fields=None):
        return self.request_json('GET', subpath, fields)

    def delete(self, subpath, fields=None):
        return self.request_json('DELETE', subpath, fields)

    def put(self, subpath, body, fields=None):
        encoded_data = encode_json(body)
        if fields is not None:
            return self.request_json('PUT', subpath + '?' + urlencode(fields), body=encoded_data)
        else:
            return self.request_json('PUT', subpath, body=encoded_data)

    def data_upload(self, data=[], **kwargs):
        """
        Data upload function, **kwargs are optional, they are encoded in the url query params

        :param repo: The repo name.
        :type repo: ``string``
        :param host: The host of data comes from.
        :type host: ``string``
        :param sourcetype: The sourcetype name.
        :type sourcetype: ``string``
        :param origin: The data origin, such as source file name etc.
        :type origin: ``string``
        :param collectTime: The collect time of the data
        :type collectTime: ``integer`` of 13 digit integer.
        :returns: response data.
        """
        return self.post(PATH_DATA_UPLOAD, data, fields=kwargs)

    def data_upload_strictly(self, data=[], **kwargs):
        """
        Encapsulate ``data_upload`` function, and check if the data is legal
        """
        repo = kwargs.get("repo", None)
        sourcetype = kwargs.get("sourcetype", None)
        i = 0
        for d in data:
            if 'raw' not in d:
                raise IllegalArgument(
                    "The {}th data does not have raw field".format(i))
            if not repo and 'repo' not in d:
                raise IllegalArgument(
                    "The {}th data does not have repo field".format(i))
            if not sourcetype and 'sourcetype' not in d:
                raise IllegalArgument(
                    "The {}th data does not have sourcetype field".format(i))
            i += 1
        return self.data_upload(data, **kwargs)

    def create_query_job(self, spl, **kwargs):
        """
        SPL Query Functions. ``create_query`` create a spl query job


        :param spl: The spl query command, such as `* | stats count()`
        :type spl: ``string``
        :param start: The search start time.
        :type start: ``integer`` of 13 digit integer.
        :param end: The search end time.
        :type end: ``integer`` of 13 digit integer.
        :param mode: The searach mode: legal options are ``fast``, ``smart`` and ``detailed``.
        :type mode: ``string``
        :param preview: When preview is True, then return intermediate results even if the query did not finish
        :type preview: ``boolean``
        :param collectSize: Limit the returned data size
        :type collectSize: ``integer``
        :returns: response data.
        """
        if len(spl) <= 0:
            raise IllegalArgument("SPL is required when create query job !")
        query = {
            "query": spl,
            "startTime": int(kwargs.get("start", 0)),
            "endTime": int(kwargs.get("end", time.time() * 1000)),
            "mode": kwargs.get("mode", "fast"),
            "preview": kwargs.get("preview", False),
            "collectSize": int(kwargs.get("collectSize", -1))
        }
        return self.post(PATH_SPL_QUERY_JOB, query)

    def get_query_status(self, query_id):
        return self.get(PATH_SPL_QUERY_JOB_STATUS.format(query_id))

    def get_query_results(self, query_id):
        return self.get(PATH_SPL_QUERY_JOB_RESULTS.format(query_id))

    def get_query_events(self, query_id):
        return self.get(PATH_SPL_QUERY_JOB_EVENTS.format(query_id))

    def get_query_timeline(self, query_id):
        return self.get(PATH_SPL_QUERY_JOB_TIMELINE.format(query_id))

    def get_query_summary(self, query_id):
        return self.get(PATH_SPL_QUERY_JOB_SUMMARY.format(query_id))

    def create_repo(self, repo_name, **kwargs):
        """
        Repo creation API with validation

        :param repo_name: The repositories' name
        :type repo_name: ``string``
        :param description: The repositories' description
        :type description: ``string``
        :param retention: The data's retention, -1 means forever
        :type retention: ``integer``
        :param lifeCycleEnable: Enable life cycle config
        :type lifeCycleEnable: ``boolean``
        :param lifeCycle: Life cycle policy, map with key hot, warm, cold maxages
        :type lifeCycle: ``map``
        :param rollover: Rollover policy, map with key shardMaxDocs, shardMaxSize, indexMaxAge
        :type rollover: ``map``
        :param replicas: Replicas policy, map with key repoReplicas, writeReplicas, writeShards, writeRefreshIntervalInSeconds
        :type replicas: ``map``
        """
        builder = RepoConfigBuilder(repo_name)
        builder.set_desc(kwargs.get("description", "")) \
            .set_retention(int(kwargs.get("retention", -1)))

        life_cycle_enable = kwargs.get("lifeCycleEnable", False)
        life_cycle = kwargs.get("lifeCycle", {})
        rollover = kwargs.get("rollover", {})
        replicas = kwargs.get("replicas", {})

        if len(rollover) > 0:
            builder.set_rollover(rollover)

        builder.set_repo_replicas(replicas.get("repoReplicas")) \
            .set_write_replicas(replicas.get("writeReplicas")) \
            .set_write_shards(replicas.get("writeShards")) \
            .set_write_refresh_interval(replicas.get("writeRefreshIntervalInSeconds"))

        if life_cycle_enable:
            builder.enable_lifecycle() \
                .init_lifecycle() \
                .set_lifecycle_ages(life_cycle)

        req_body = builder.build()
        return self.create_repo_by_body(repo_name, req_body)

    def get_repos(self, **page_params):
        """
        Get repo list

        :param sort: The sort column, updateTime by default
        :type sort: ``string``
        :param order: The order of data, asc or desc, desc by default
        :type order: ``string``
        :param pageNo: The page no, start from 1
        :type pageNo: ``integer``
        :param pageSize: The size of page, 10 by default
        :type pageSize: ``integer``
        :param prefix: The search keyword of repo name
        :type prefix: ``string``
        """
        return self.get(PATH_REPOS, page_params)

    def get_repo_by_name(self, repo_name):
        return self.get(PATH_SINGLE_REPO.format(repo_name))

    def delete_repo_by_name(self, repo_name):
        return self.delete(PATH_SINGLE_REPO.format(repo_name))

    def create_repo_by_body(self, repo_name, req_body):
        return self.post(PATH_SINGLE_REPO.format(repo_name), req_body)

    def update_repo_by_body(self, repo_name, req_body):
        return self.put(PATH_SINGLE_REPO.format(repo_name), req_body)


def encode_json(data):
    return json.dumps(data).encode(DEFAULT_ENCODING)


def decode_json(data_bytes):
    return json.loads(data_bytes.decode(DEFAULT_ENCODING))



import json
from ..storage.base import (Base, UrlEncoded)

__all__ = [
    "connect",
    "Service"
]

def _path(base, name):
    if not base.endswith('/'): base = base + '/'
    return base + name

# kwargs: scheme, host, port, app
def connect(**kwargs):
    s = Service(**kwargs)
    return s

class Service(Base):
    """
    :param host: Host name. type ``string``
    :param port: Port number. type ``integer``
    :param scheme: Scheme for accessing the service. type "http" or "https"
    :return: A :class:`Service` instance.
    """
    def __init__(self, **kwargs):
        super(Service, self).__init__(**kwargs)
        self._ml_version = None

    def storage(self, app):
        """
        :param app: App name. type ``string``
        Return KV Store table.
        :return: A :class:`StorageTable`.
        """
        return StorageTable(self, app)


class Endpoint(object):
    """
    An ``Endpoint`` represents a URI.
    This class provides :class:`Entity` (HTTP GET and POST methods).
    """
    def __init__(self, service, path):
        self.service = service
        self.path = path if path.endswith('/') else path + '/'

    def get(self, path_segment="", **kwargs):
        """
        A GET operation on the path segment.
        """
        if path_segment.startswith('/'):
            path = path_segment
        else:
            path = self.service._abspath(self.path + path_segment)
        return self.service.get(path, **kwargs)

    def post(self, path_segment="", body=None):
        """
        A POST operation on the path segment.
        """
        if path_segment == "":
            path = self.path
        elif path_segment.startswith('/'):
            path = path_segment
        else:
            path = self.service._abspath(self.path + path_segment)
        return self.service.post(path, None, body=body)
    
    def put(self, path_segment="", body=None):
        """
        A PUT operation on the path segment.
        """
        if path_segment == "":
            path = self.path
        elif path_segment.startswith('/'):
            path = path_segment
        else:
            path = self.service._abspath(self.path + path_segment)
        return self.service.put(path, None, body=body)    
    
    def delete(self, path_segment="", **kwargs):
        """
        A Delete operation on the path segment.
        """
        if path_segment == "":
            path = self.path
        elif path_segment.startswith('/'):
            path = path_segment
        else:
            path = self.service._abspath(self.path + path_segment)
        return self.service.delete(path, **kwargs)    
    
class StorageTable(Endpoint):
    def __init__(self, service, app):
        Endpoint.__init__(self, service, '/api/v1/storage/collections/' + UrlEncoded(app))
        self.app = app
        
    def data(self, name):
        """
        Return data object for this Collection. rtype: :class:`KVStoreCollectionData`
        """
        return StorageTableData(self, self.app, name)

    def create(self, data):
        """
        Create a KV Store table.

        :param app: App name. type ``string``
        :param name: Table name. type ``string``
        :param schema: Table schema. type ``dict``

        :return: Result of POST request
        """
        return self.post("config", body=data)
    
    def get_tables(self):
        """
        Get a KV Store table.

        :param name: Table name. type ``string``

        :return: Result of POST request
        """
        return json.loads(self.get("config").body)
    
    def delete_table(self, name):
        """
        Get a KV Store table.

        :param name: Table name. type ``string``

        :return: Result of POST request
        """
        return json.loads(self.delete(name).body)     

class StorageTableData(object):
    """
    Represent the data endpoint for a StorageTable. Using :meth:`StorageTable.data`
    """
    def __init__(self, service, app, name):
        self.service = service
        self.path = '/api/v1/storage/collections/' + app +  "/data/" +  name

    def _get(self, url, **kwargs):
        return self.service.get(self.path + url, **kwargs)

    def _post(self, url, body):
        return self.service.post(self.path + url, body=body)
    
    def _put(self, url, body):
        return self.service.put(self.path + url, body=body)    

    def _delete(self, url, **kwargs):
        return self.service.delete(self.path + url, **kwargs)

    def query(self, **kwargs):
        """
        Get the results of query.

        :param kwargs: Parameters (Optional). Such as sort, limit, skip, and fields. type ``dict``
        :return: Array of documents. rtype: ``array``
        """
        return json.loads(self._get('', **kwargs).body)

    def query_by_id(self, id):
        """
        Return object with id.

        :param id: Value for ID.
        :return: Document with id. rtype: ``dict``
        """
        return json.loads(self._get("/" + str(id)).body)

    def insert(self, record):
        """
        Insert item into this table. An id field will be generated in the data.

        :param data: Document to insert. type ``string``
        :return: id of inserted object. rtype: ``dict``
        """
        data = json.dumps(record)
        return json.loads(self._post('', data).body)
    
    def delete(self, **kwargs):
        """
        Delete.

        :param kwargs: Parameters (Optional). Such as sort, limit, skip, and fields. type ``dict``
        :return: Result of DELETE request
        """
        return json.loads(self._delete('', **kwargs).body)    

    def delete_by_id(self, id):
        """
        Delete by id.

        :param id: id record to delete. type ``string``
        :return: Result of DELETE request
        """
        return json.loads(self._delete("/" + str(id)).body)

    def update(self, id, data):
        """
        Replace record with id and data.

        :param id: Id of record to update. type ``string``
        :param data: The new record to insert. type ``string``
        :return: id of replaced record`
        """
        return json.loads(self._put("/" + str(id), body=data).body)
    
    def updateByQuery(self, data):
        """
        Replace record with id and data.

        :param id: Id of record to update. type ``string``
        :param data: The new record to insert. type ``string``
        :return: id of replaced record`
        """
        return json.loads(self._put("", body=data).body)    

    def batch_save(self, records):
        """
        Insert records in records.

        :param records: Array of records to save as dictionaries. type ``array`` of ``dict``
        :return: Results of insert Request.
        """
        if len(records) < 1:
            raise Exception('Must have at least one record.')

        data = json.dumps(records)
        return json.loads(self._post('/batch_save', body=data).body)




#!/usr/bin/env python
import sys

from pdr_python_sdk.api.on_demand_api import OnDemandApi
from pdr_python_sdk.api.response import Response
from pdr_python_sdk.on_demand_action import run

class HelloWorldApi(OnDemandApi):
    def do_handle_data(self, data):
        if not data.contains_request():
            raise Exception('api data should contain request details')
        request = data.request()
        if 'GET' != str.upper(request.method()):
            return Response(405, 'unsupported method [{}]'.format(str.upper(request.method()))).to_string()
        return Response(204, [{"hello":"world"}]).to_string()


if __name__ == '__main__':
    run(HelloWorldApi, sys.argv, sys.stdin.buffer, sys.stdout.buffer)




class Record(dict):
    """
    Enable dot access to members of a dictionary.
    """
    sep = '.'

    def __call__(self, *args):
        if len(args) == 0: 
            return self
        return Record((key, self[key]) for key in args)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError: 
            raise AttributeError(name)

    def __delattr__(self, name):
        del self[name]

    def __setattr__(self, name, value):
        self[name] = value

    @staticmethod
    def fromkv(k, v):
        result = record()
        result[k] = v
        return result

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        key += self.sep
        result = record()
        for k,v in iter(self.items()):
            if not k.startswith(key):
                continue
            suffix = k[len(key):]
            if '.' in suffix:
                ks = suffix.split(self.sep)
                z = result
                for x in ks[:-1]:
                    if x not in z:
                        z[x] = record()
                    z = z[x]
                z[ks[-1]] = v
            else:
                result[suffix] = v
        if len(result) == 0:
            raise KeyError("No key or prefix: %s" % key)
        return result
    

def record(value=None): 
    """
    Return a :class:`Record` instance with value provided.
    :param `value`: An initial record value. type ``dict``
    """
    if value is None: 
        value = {}
    return Record(value)


import sys

from .spl_base_command import SplBaseCommand


class SplStreamingBatchCommand(SplBaseCommand):
    def process_data(self, parser, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        while True:
            execute_meta = self.process_protocol_execute(input_stream, parser)
            resp = self.streaming_handle(self.lines)
            parser.send_packet(output_stream, execute_meta, resp)
            self.lines = []
            if self.is_finish:
                break



import yaml
from pdr_python_sdk.client import *

"""
Demo connection params

params = {
    "scheme": "http",
    "host": "pandora-express-rc.qiniu.io",
    "port": 80,
    "token": "xxxxx"
}
"""
config_file = open("config.yaml", 'r')
params = yaml.load(config_file, Loader=yaml.FullLoader)
conn = connect(**params)
config_file.close()


def data_upload_demo():
    """
    上传接口demo
    """
    data = [
        # 全字段的数据
        {
            "raw": """{"a":"de", "dd":1}""",
            "repo": "default",
            "sourcetype": "json",
            "host": "192.168.1.1",
            "origin": "/path/to/file/log.txt",
            "collectTime": 1582619883000,
            "timestamp": 1582619883000
        },
        # 如果repo和sourcetype 全局制指定了，就可以在数据中忽略掉
        {
            "raw": """{"a":"de", "dd":1}""",
            "host": "192.168.1.1",
            "origin": "/path/to/file/log.txt",
            "collectTime": 1582619883000,
            "timestamp": 1582619883000
        },
        # 最简单的数据上传
        {
            "raw": """{"a":"de", "dd":1}"""
        }
    ]
    # 上传前会做参数检验
    conn.data_upload_strictly(data=data, repo="default", sourcetype="json")
    # 上传前不会做参数检验
    conn.data_upload(data=data, repo="default", sourcetype="json")


def search_events_demo():
    """
    搜索型SPL demo
    """
    # 查看前一小时前端错误
    # fast 模式只计算内置字段
    # smart 模式只计算搜索的数据中出现的字段
    # detailed 模式计算所有字段的统计信息
    query_info = conn.create_query_job(
        """ repo=_frontend_error """,
        start=int(time.time() * 1000 - 60 * 60 * 1000),
        end=int(time.time() * 1000),
        mode="smart",
        collectSize=1000
    )
    # 从返回结果中获得query_id
    query_id = query_info['id']
    # 轮询获得搜索任务是否成功
    while True:
        last_status = conn.get_query_status(query_id)
        if last_status['process'] == 1:
            break
        print("current status is {} ".format(last_status))
        time.sleep(0.1)
    print("final status is {} ".format(last_status))
    # 搜索型SPL
    assert not last_status['isResult']
    events = conn.get_query_events(query_id)
    print("events count is {} ".format(events['total']))
    print("events rows are {} ".format(events['rows']))

    timeline = conn.get_query_timeline(query_id)
    print("timeline buckets are {} ".format(timeline['buckets']))

    summary = conn.get_query_summary(query_id)
    print("summary of fields statistics info is {}".format(summary))


def search_stats_demo():
    """
    计算型搜索demo
    """
    # 查看前一小时前端错误
    # fast 模式只计算内置字段
    # smart 模式只计算搜索的数据中出现的字段
    # detailed 模式计算所有字段的统计信息
    query_info = conn.create_query_job(
        """ repo=_frontend_error | stats count() as cnt""",
        start=int(time.time() * 1000 - 60 * 60 * 1000),
        end=int(time.time() * 1000),
        mode="fast",
        collectSize=1000
    )
    # 从返回结果中获得query_id
    query_id = query_info['id']
    # 轮询获得搜索任务是否成功
    while True:
        last_status = conn.get_query_status(query_id)
        if last_status['process'] == 1:
            break
        print("current status is {} ".format(last_status))
        time.sleep(0.1)
    print("final status is {} ".format(last_status))
    # 计算型SPL
    assert last_status['isResult']
    results = conn.get_query_results(query_id)
    print("result fields are {} ".format(results['fields']))
    print("result rows are {} ".format(results['rows']))

def repo_management_demo():
    """
    仓库管理相关demo，不使用分层存储
    """
    # 创建仓库。不使用分层存储
    conn.create_repo("testrepo_by_sdk", lifeCycleEnable=False)

    # 获得仓库列表
    repos = conn.get_repos(prefix="testrepo_by_sdk", pageSize=5, pageNo=1)
    print(repos)

    # 获取单个仓库配置
    config = conn.get_repo_by_name("testrepo_by_sdk")

    config["description"] = "to update by sdk"

    # 如果不启动分层存储，需要手动把分层策略从配置中删除
    if not config['lifecyclePolicyEnable']:
        config.pop("lifecyclePolicy")

    # 更新仓库配置
    conn.update_repo_by_body("testrepo_by_sdk", config)
    # 删除仓库
    conn.delete_repo_by_name("testrepo_by_sdk")


    # 创建仓库。使用分层存储，分层存储时间单位为毫秒
    # 7天热存储
    # 30天温存储
    # 365天冷存储
    conn.create_repo("testlayer_by_sdk", lifeCycleEnable=True, lifeCycle={
        "hot": 7*24*60*60*1000,
        "warm": 30*24*60*60*1000,
        "cold": 365*24*60*60*1000
    })
    conn.delete_repo_by_name("testlayer_by_sdk")

if __name__ == "__main__":
    repo_management_demo()



import sys

from .packet import ApiPacketBody
from .packet import ApiRequestPacket
from .packet import ApiResponsePacket
from ..on_demand_action import OnDemandAction


class OnDemandApi(OnDemandAction):
    """
    One time api which process starts on new request and ends after handling it.
    """

    def __init__(self):
        self.__is_inited = False
        self.__input_stream = sys.stdin.buffer
        self.__output_stream = sys.stdout.buffer

    def do_handle_init(self, packet):
        """
        actual initialization actions
        @param ApiRequestPacket packet: packet received from the request
        """
        pass

    def do_handle_data(self, data):
        """
        actual data processing actions
        @param ApiPacketBody data: data block in request
        @rtype: str
        @return: response as a string
        """
        raise NotImplemented('method [do_handle_data(data)] is not yet implemented')

    def do_handle_end(self):
        """
        perform terminating processes
        @return:
        """
        pass

    def handle_init(self, packet):
        """
        load api according to arrived request, for example, setup session, assign env variables, etc.
        @param ApiRequestPacket packet: packet received from the request
        """
        self.do_handle_init(packet)
        self.__is_inited = True

    def handle_data(self, data):
        """
        actions to handle packet body and generate response
        @param ApiPacketBody data: data block in request
        @rtype: str
        @return: response as a string
        """
        return self.do_handle_data(data)

    def handle_end(self):
        """
        perform actions needed after request is handled, for example, release resources, clear thread, etc.
        """
        self.do_handle_end()

    def write(self, data):
        """
        write response to output stream
        @param str data: response string
        """
        self.__output_stream.write(data.encode('utf-8'))

    def write_packet(self, packet):
        """
        write packet to output stream
        @param packet: packet to be writen to output stream
        """
        self.write(packet.to_string())

    def on_request(self, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        """
        default method called when new request arrives.
        in default, this method uses sys.stdin as input stream and sys.stdout as ouput stream
        """
        self.handle_request(argv, input_stream, output_stream)

    def handle_request(self, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        """
        handle request from given input_stream, and response through given output_stream
        @param argv: cmd arguments
        @param input_stream: input_stream to read request
        @param output_stream: output_stream to write response
        """
        self.__input_stream = input_stream
        self.__output_stream = output_stream

        packet = ApiRequestPacket()
        packet.read(self.__input_stream)

        if packet.is_init():
            try:
                if self.__is_inited:
                    raise ApiAlreadyInitedException('current api has already been initialized')
                self.handle_init(packet)
            except:
                self.write('error when handling init request')
                raise
            # write empty message, which means initialization succeeded
            self.write('')

        if packet.contains_data():
            try:
                response = self.handle_data(packet.body())
                self.write_packet(ApiResponsePacket(1, response))
                pass
            except:
                self.write('error when handling data')
                raise

        if packet.is_end():
            self.handle_end()


class OnDemandApiException(Exception):
    pass


class ApiAlreadyInitedException(OnDemandApiException):
    pass



import logging
import sys
import traceback

from .spl_packet_utils import SplPacketUtils
from ..on_demand_action import OnDemandAction


class SplBaseCommand(OnDemandAction):
    def __init__(self):
        self.uri = 'http://127.0.0.1:9999'
        self.session = ''
        self.metainfo = None
        self.is_finish = False
        self.require_fields = ['*']
        self.export_fields = []
        self.lines = []
        self.spl_args = []

    def on_request(self, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        self.process_protocol(argv, input_stream, output_stream)

    def process_protocol(self, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        try:
            if argv is None:
                argv = sys.argv
            logging.debug('execute script command: {}'.format(argv))
            parser = SplPacketUtils()
       
            self.process_protocol_info(input_stream, parser)
            self.init_env_by_getinfo()
            self.metainfo['require_fields'] = self.config_require_fields()
            self.metainfo['export_fields'] = self.config_export_fields()
            parser.send_packet(output_stream, self.metainfo, [])
            self.after_getinfo()
            self.process_data(parser, argv, input_stream, output_stream)
        except Exception as error:
            logging.exception(error)
            self.metainfo['error_message'] = "{}".format(error)
            self.metainfo['error_traceback'] = "{}".format(traceback.format_exc())
            parser.send_packet(output_stream, self.metainfo, [])

    def process_data(self, parser, argv=None, input_stream=sys.stdin.buffer, output_stream=sys.stdout.buffer):
        return

    def process_protocol_info(self, input_stream, parser):
        meta_length, body_length = parser.parse_head(input_stream)
        if meta_length <= 0:
            raise RuntimeError('GetInfo Protocol metaLength is invalid: {}'.format(meta_length))

        self.metainfo = parser.parse_meta(input_stream, meta_length)

        # discard body in getinfo
        parser.parse_body(input_stream, body_length)

    def process_protocol_execute(self, input_stream, parser):
        meta_length, body_length = parser.parse_head(input_stream)
        if meta_length <= 0:
            raise RuntimeError('Execute Protocol metaLength is invalid: {}'.format(meta_length))

        execute_meta = parser.parse_meta(input_stream, meta_length)

        if execute_meta['action'] != "execute":
            raise RuntimeError('Execute Protocol action is invalid: {}'.format(execute_meta['action']))

        self.is_finish = execute_meta['finished']
        tmp = parser.parse_body(input_stream, body_length)
        if len(tmp) > 0:
            self.lines.extend(tmp)
        return execute_meta

    def after_getinfo(self):
        return

    def init_env_by_getinfo(self):
        return

    def config_require_fields(self, require_fields=None):
        if require_fields is None:
            require_fields = []
        if require_fields is None or len(require_fields) == 0:
            require_fields = ['*']
        return require_fields

    def config_export_fields(self, export_fields=None):
        if export_fields is None:
            export_fields = []
        if export_fields is None or len(export_fields) == 0:
            export_fields = ['*']
        return export_fields

    def streaming_handle(self, lines):
        return lines, True



import json
import sys

class SplPacketUtils(object):
    def parse_head(self, input_stream=sys.stdin.buffer):
        try:
            header = input_stream.readline().decode("utf-8")
        except Exception as error:
            raise RuntimeError('Failed to read spl protocol header: {}'.format(error))
        parts = str.split(header, ",")
        meta_length = int(parts[1])
        body_length = int(parts[2])

        return meta_length, body_length

    def parse_meta(self, input_stream=sys.stdin.buffer, length=0):
        try:
            meta_body = input_stream.read(length).decode("utf-8")
            metainfo = json.loads(meta_body)
        except Exception as error:
            raise RuntimeError('Failed to parser spl protocol meta: {}'.format(error))

        return metainfo

    def parse_body(self, input_stream=sys.stdin.buffer, length=0):
        records = []
        if length <= 0:
            return records

        try:
            body = input_stream.read(length).decode("utf-8")
            rows = str.split(body, '\n')
            if len(rows) < 2:
                return records

            fields = str.split(rows[0], '\t')
            for row in rows[1:]:
                record = {}
                parts = str.split(row, '\t')
                for i in range(len(fields)):
                    if i < len(parts):
                        record[fields[i]] = self.format_field(parts[i])
                records.append(record)

            return records
        except Exception as error:
            raise RuntimeError('Failed to parser spl protocol body: {}'.format(error))

    def format_field(self, part):
        type, value = str.split(part, ',', 1)
        if int(type) == 0:
            return self.decode_string(value)
        elif int(type) == 1:
            return int(value)
        elif int(type) == 2:
            return float(value)
        elif int(type) == 3:
            return ''
        elif int(type) == 4:
            return str.islower(value) == 'true'
        elif int(type) == 5:
            return value
        return value


    def send_packet(self, output_stream=sys.stdout.buffer, meta_info=None, lines=None):
        if meta_info is None:
            meta_info = {}
        if lines is None:
            lines = []
        meta = json.dumps(meta_info).encode("utf-8")

        body = self.convert_body_to_str(lines).encode("utf-8")

        head = ('chunked 1.0,%s,%s\n' % (len(meta), len(body))).encode("utf-8")
        output_stream.write(head)
        output_stream.write(meta)
        if len(body) > 0:
            output_stream.write(body)

        output_stream.flush()
        return

    def convert_body_to_str(self, lines=[]):
        if lines is None:
            return ''

        if len(lines) == 0:
            return ''

        field_str = ""
        line_strs = []
        fields = []
        allfields = {}
        for line in lines:
            if not isinstance(line, dict):
                continue
            row = dict(line)
            line_str = ''
            if len(fields) != 0:
                for field in fields:
                    if field not in row:
                        line_str += '\t'
                    else:
                        line_str = line_str + self.encode_string(str(row[field])) + '\t'

            for key in row:
                if key in allfields:
                    continue
                allfields[key] = 0
                fields.append(key)
                line_str = line_str + self.encode_string(str(row[key])) + '\t'

            line_strs.append(line_str[:len(line_str) - 1])

        for field in fields:
            field_str = field_str + field + '\t'

        field_str = field_str[:len(field_str) - 1]

        body = field_str
        for line_str in line_strs:
            body += '\n' + line_str

        return body

    def encode_string(self, value):
        value = str.replace(value, '\t', '\\t')
        value = str.replace(value, '\n', '\\n')
        return value

    def decode_string(self, value):
        value = str.replace(value, '\\t', '\t')
        value = str.replace(value, '\\n', '\n')
        return value



from setuptools import setup, find_packages

setup(
    name = "pdr-python-sdk",
    version = "1.0.4",
    keywords = ["pip", "qiniu", "pandora", "sdk"],
    description = "pandora python sdk, include customapi, customoperation, kvstore",
    license = "MIT Licence",

    author = "qiniu",
    author_email = "pandora@qiniu.com",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = [
        "urllib3",
        "pyyaml"
    ]
)


import json
import logging


class ApiRequestPacket(object):
    """
    packet used to hold api request data
    """

    OPCODE_REQUEST_INIT = 0x01
    OPCODE_REQUEST_DATA = 0x02
    OPCODE_REQUEST_END = 0x04
    OPCODE_REQUEST_ALLOW_STREAM = 0x08

    def __init__(self):
        self.__opcode = None
        self.__body_length = None
        self.__body = None

    def is_init(self):
        """
        returns whether a packet is an init packet
        @rtype: bool
        @return: returns whether a packet is an init packet
        """
        return (self.__opcode & ApiRequestPacket.OPCODE_REQUEST_INIT) != 0

    def is_end(self):
        """
        returns whether a packet is an end packet
        @rtype: bool
        @return: returns whether a packet is an end packet
        """
        return (self.__opcode & ApiRequestPacket.OPCODE_REQUEST_END) != 0

    def contains_data(self):
        """
        returns whether a packet contains data
        @rtype: bool
        @return: returns whether a packet contains data
        """
        return (self.__opcode & ApiRequestPacket.OPCODE_REQUEST_DATA) != 0

    def allow_stream(self):
        """
        returns whether streaming is allowed
        @rtype: bool
        @return: returns whether a packet allows streaming
        """
        return (self.__opcode & ApiRequestPacket.OPCODE_REQUEST_ALLOW_STREAM) != 0

    def opcode(self):
        """
        returns opcode of the packet
        @rtype: int
        @return: packet opcode
        """
        return self.__opcode

    def body_length(self):
        """
        returns packet body length
        @rtype: int
        @return: packet body length
        """
        return self.__body_length

    def body(self):
        """
        returns request received
        @rtype: ApiPacketBody
        @return: request
        """
        return self.__body

    def read(self, input_stream):
        """
        read request packet from input stream
        @param input_stream: input stream that request data will be read from
        """
        while True:
            opcode_str = input_stream.readline().decode('utf-8')
            if opcode_str == b'':
                return False
            if opcode_str != b'\n':
                # ignore extra newlines before opcode
                break
            break
        self.__opcode = int(opcode_str)
        len_str = input_stream.readline().decode("utf-8")
        self.__body_length = int(len_str)
        request_string = input_stream.read(self.__body_length).decode("utf-8")
        self.__body = _parse_body(request_string)

        return True


class ApiResponsePacket(object):
    """
    api response packet contains response opcode and response body
    """

    def __init__(self, opcode, body):
        self.__opcode = opcode
        self.__body = body

    def opcode(self):
        return self.__opcode

    def body_length(self):
        return len(self.__body)

    def body(self):
        return self.__body

    def to_string(self):
        """
        convert a response packet to a string, which
        @return:
        """
        return str(self.opcode()) + '\n' + str(self.body_length()) + '\n' + self.body()


def _parse_body(body):
    return ApiPacketBody(**json.loads(body))


class ApiPacketBody(object):
    def __init__(self, metadata, request):
        if metadata is None:
            self.__metadata = None
        else:
            self.__metadata = metadata

        if request is None:
            self.__request = None
        else:
            self.__request = ApiRequest(**request)

    def contains_metadata(self):
        return self.__metadata is not None

    def contains_request(self):
        return self.__request is not None

    def metadata(self):
        return self.__metadata

    def request(self):
        return self.__request


class ApiRequest(object):
    def __init__(self, method=None, header=None, param=None, path=None, requestBody=None, uuid=None):
        self.__method = method
        self.__header = header
        self.__param = param
        self.__path = path
        self.__request_body = requestBody
        self.__uuid = uuid

    def method(self):
        return self.__method

    def header(self):
        return self.__header

    def param(self):
        return self.__param

    def path(self):
        return self.__path

    def request_body(self):
        return self.__request_body

    def uuid(self):
        return self.__uuid



from __future__ import absolute_import

import logging
import socket
import ssl

import urllib
import http.client
from ..storage.record import record


__all__ = [
    "connect",
    "Base",
    "handler",
    "HTTPError"
]

# If you change these, update the docstring
# on _uri as well.
DEFAULT_HOST = "localhost"
DEFAULT_PORT = "9200"
DEFAULT_SCHEME = "http"


class UrlEncoded(str):
    """
    URL-encoded strings.

    ``UrlEncoded`` objects are identical to ``str`` objects (including being
    equal if their contents are equal) except when passed to ``UrlEncoded``
    again.

    ``UrlEncoded`` removes the ``str`` type support for interpolating values
    with ``%`` (doing that raises a ``TypeError``). There is no reliable way to
    encode values this way, so instead, interpolate into a string, quoting by
    hand, and call ``UrlEncode`` with ``skip_encode=True``.
    """
    def __new__(self, val='', skip_encode=False, encode_slash=False):
        if isinstance(val, UrlEncoded):
            return val
        elif skip_encode:
            return str.__new__(self, val)
        elif encode_slash:
            return str.__new__(self, urllib.parse.quote_plus(val))
        else:
            return str.__new__(self, urllib.parse.quote(val))

    def __add__(self, other):
        """
        self + other

        If *other* is not a ``UrlEncoded``, URL encode it before
        adding it.
        """
        if isinstance(other, UrlEncoded):
            return UrlEncoded(str.__add__(self, other), skip_encode=True)
        else:
            return UrlEncoded(str.__add__(self, urllib.parse.quote(other)), skip_encode=True)

    def __radd__(self, other):
        """
        other + self

        If *other* is not a ``UrlEncoded``, URL _encode it before
        adding it.
        """
        if isinstance(other, UrlEncoded):
            return UrlEncoded(str.__radd__(self, other), skip_encode=True)
        else:
            return UrlEncoded(str.__add__(urllib.parse.quote(other), self), skip_encode=True)

    def __mod__(self, fields):
        """
        Interpolation into ``UrlEncoded``s is disabled.

        If you try to write ``UrlEncoded("%s") % "abc", will get a
        ``TypeError``.
        """
        raise TypeError("Cannot interpolate into a UrlEncoded object.")
    def __repr__(self):
        return "UrlEncoded(%s)" % repr(urllib.parse.unquote(str(self)))

def _uri(scheme=DEFAULT_SCHEME, host=DEFAULT_HOST, port=DEFAULT_PORT):
    """
    Construct a URI from the given *scheme*, *host*, and *port*.

    :param scheme: URL scheme (the default is "https")
    :type scheme: "http" or "https"
    :param host: The host name (the default is "localhost")
    :type host: string
    :param port: The port number (the default is 8089)
    :type port: integer
    :return: The URI.
    :rtype: UrlEncoded (subclass of ``str``)
    """
    if ':' in host:
        # IPv6 addresses must be enclosed in [ ] in order to be well
        # formed.
        host = '[' + host + ']'
    return UrlEncoded("%s://%s:%s" % (scheme, host, port), skip_encode=True)

class Base(object):
    """
    The ``Base`` class encapsulates the details of HTTP requests,
    authentication, and URL prefixes to simplify access to
    the REST API.

    :param host: The host name (the default is "localhost").
    :type host: ``string``
    :param port: The port number (the default is 8089).
    :type port: ``integer``
    :param scheme: The scheme for accessing the service (the default is "https").
    :type scheme: "https" or "http"
    :param headers: List of extra HTTP headers to send (optional).
    :type headers: ``list`` of 2-tuples.
    :param handler: The HTTP request handler (optional).
    :returns: A ``Base`` instance.
    """
    def __init__(self, handler=None, **kwargs):
        self.http = HttpBase(handler, kwargs.get("verify", False), key_file=kwargs.get("key_file"),
                            cert_file=kwargs.get("cert_file"))  # Default to False for backward compat
        self.scheme = kwargs.get("scheme", DEFAULT_SCHEME)
        self.host = kwargs.get("host", DEFAULT_HOST)
        self.port = int(kwargs.get("port", DEFAULT_PORT))
        self.uri = _uri(self.scheme, self.host, self.port)
        self.additional_headers = kwargs.get("headers", [])
        self.additional_headers = self.assign_default_header(self.additional_headers, 'Content-Type', 'application/json')

    def assign_default_header(self, headers, header, value):
        exists = False
        for h in headers:
            if h[0] == header:
                exists = True
                break

        if not exists:
            return headers + [(header, value)]


    def connect(self):
        """
        Return an open connection (socket).

        This method is used for writing bulk events to an index or similar tasks
        where the overhead of opening a connection multiple times would be
        prohibitive.

        :returns: A socket.
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.scheme == "https":
            sock = ssl.wrap_socket(sock)
        sock.connect((socket.gethostbyname(self.host), self.port))
        return sock

    def delete(self, path_segment, **kwargs):
        """
        Perform a DELETE operation at the REST path segment.

        This method is named to match the HTTP method.

        :raises HTTPError: Raised when an error occurred in a GET operation from
             *path_segment*.
        :param path_segment: A REST path segment. type ``string``
        :param kwargs: All other arguments. type ``dict``
        :return: The response from the server.
        """
        path = self.uri + self._abspath(path_segment)
        logging.debug("DELETE request to %s (body: %s)", path, repr(kwargs))
        response = self.http.delete(path, self.additional_headers, **kwargs)
        return response

    def get(self, path_segment, headers=None, **kwargs):
        """
        Perform a GET operation from the REST path segment.

        :raises HTTPError: Raised when an error occurred in a GET operation from
             *path_segment*.
        :param path_segment: A REST path segment. type ``string``
        :param headers: List of extra HTTP headers to send (optional). type ``list`` of 2-tuples.
        :param kwargs: All other arguments. type ``dict``
        :return: The response from the server.
        """
        if headers is None:
            headers = []

        path = self.uri + self._abspath(path_segment)
        logging.debug("GET request to %s (body: %s)", path, repr(kwargs))
        all_headers = headers + self.additional_headers
        response = self.http.get(path, all_headers, **kwargs)
        return response

    def post(self, path_segment, headers=None, **kwargs):
        """
        Perform a POST operation from the REST path segment.

        :raises HTTPError: Raised when an error occurred in a GET operation from *path_segment*.
        :param path_segment: A REST path segment. type ``string``
        :param headers: List of extra HTTP headers (optional). type ``list`` of 2-tuples.
        :param kwargs: All other arguments. type ``dict``
        :return: The response from the server.
        """
        if headers is None:
            headers = []

        path = self.uri + self._abspath(path_segment)
        logging.debug("POST request to %s (body: %s)", path, repr(kwargs))
        all_headers = headers + self.additional_headers
        response = self.http.post(path, all_headers, **kwargs)
        return response
    
    def put(self, path_segment, headers=None, **kwargs):
        """
        Perform a POST operation from the REST path segment.

        :raises HTTPError: Raised when an error occurred in a GET operation from *path_segment*.
        :param path_segment: A REST path segment. type ``string``
        :param headers: List of extra HTTP headers (optional). type ``list`` of 2-tuples.
        :param kwargs: All other arguments. type ``dict``
        :return: The response from the server.
        """
        if headers is None:
            headers = []

        path = self.uri + self._abspath(path_segment)
        logging.debug("PUT request to %s (body: %s)", path, repr(kwargs))
        all_headers = headers + self.additional_headers
        response = self.http.put(path, all_headers, **kwargs)
        return response    

    def request(self, path_segment, method="GET", headers=None, body=""):
        """
        Issue an arbitrary HTTP request to the REST path segment.

        This method is named to match ``httplib.request``. This function
        makes a single round trip to the server.

        :raises HTTPError: Raised when an error occurred in a GET operation from
             *path_segment*.
        :param path_segment: A REST path segment.
        :type path_segment: ``string``
        :param method: The HTTP method to use (optional).
        :type method: ``string``
        :param headers: List of extra HTTP headers to send (optional).
        :type headers: ``list`` of 2-tuples.
        :param body: Content of the HTTP request (optional).
        :type body: ``string``
        :return: The response from the server.
        """
        if headers is None:
            headers = []

        path = self.uri + self._abspath(path_segment)
        all_headers = headers + self.additional_headers
        logging.debug("%s request to %s (headers: %s, body: %s)",
                      method, path, str(all_headers), repr(body))
        response = self.http.request(path,
                                     {'method': method,
                                     'headers': all_headers,
                                     'body': body})
        return response
    
    def _abspath(self, path_segment):
        skip_encode = isinstance(path_segment, UrlEncoded)
        return UrlEncoded(path_segment, skip_encode=skip_encode)

class HTTPError(Exception):
    """This exception is raised for HTTP responses that return an error."""
    def __init__(self, response, _message=None):
        status = response.status
        reason = response.reason
        body = response.body
        self.status = status
        self.reason = reason
        self.headers = response.headers
        self.body = body
        self._response = response

def _encode(**kwargs):
    items = []
    for key, value in iter(kwargs.items()):
        if isinstance(value, list):
            items.extend([(key, item) for item in value])
        else:
            items.append((key, value))
    return urllib.parse.urlencode(items)

# Split the url into (scheme, host, port, path)
def _spliturl(url):
    parsed_url = urllib.parse.urlparse(url)
    host = parsed_url.hostname
    port = parsed_url.port
    path = '?'.join((parsed_url.path, parsed_url.query)) if parsed_url.query else parsed_url.path
    # Strip brackets if its an IPv6 address
    if host.startswith('[') and host.endswith(']'): host = host[1:-1]
    if port is None: port = DEFAULT_PORT
    return parsed_url.scheme, host, port, path

# Given an HTTP request handler, this wrapper objects provides a related
# family of convenience methods built using that handler.
class HttpBase(object):
    """
    ``HttpBase`` provides `request` `delete` `post` `get` methods

    The handling function: ``handler(`url`, `request_dict`) -> response_dict``

    `url` is a dictionary with the following keys:

        - method: The method for the request, typically ``GET`` ``POST`` ``DELETE``.
        - headers: A list of pairs specifying the HTTP headers (for example: ``[('key': value), ...]``).
        - body: A string containing the body to send (default to '').

    ``response_dict`` is a dictionary with the following keys:

        - status: An integer containing the HTTP status code (such as 200 or 404).
        - reason: The reason phrase, if any, returned by the server.
        - headers: A list of pairs containing the response headers (for example, ``[('key': value), ...]``).
        - body: Response body.
    """
    def __init__(self, custom_handler=None, verify=False, key_file=None, cert_file=None):
        if custom_handler is None:
            self.handler = handler(verify=verify, key_file=key_file, cert_file=cert_file)
        else:
            self.handler = custom_handler

    def delete(self, url, headers=None, **kwargs):
        """
        Send a DELETE request to a URL.

        :param url: The URL. type ``string``
        :param headers: A list of pairs specifying the headers (for example, ``[('Content-Type', 'application/json')]``). type ``list``
        :param kwargs: Additional keyword arguments (optional). type ``dict``
        :returns: A dict (see :class:`HttpBase` for its structure).
        """
        if headers is None: headers = []
        if kwargs:
            url = url + UrlEncoded('?' + _encode(**kwargs), skip_encode=True)
        message = {
            'method': "DELETE",
            'headers': headers,
        }
        return self.request(url, message)

    def get(self, url, headers=None, **kwargs):
        """
        Send a GET request to a URL.

        :param url: The URL. type ``string``
        :param headers: A list of pairs specifying the headers for the HTTP
            response (for example, ``[('Content-Type', 'application/json')]``). type ``list``
        :param kwargs: Additional arguments (optional). type ``dict``
        :returns: A dict (see :class:`HttpBase` for its structure).
        """
        if headers is None: headers = []
        if kwargs:
            url = url + UrlEncoded('?' + _encode(**kwargs), skip_encode=True)
        return self.request(url, { 'method': "GET", 'headers': headers })

    def post(self, url, headers=None, **kwargs):
        """
        Send a POST request to a URL.

        :param url: The URL. type ``string``
        :param headers: A list of pairs specifying the headers for the HTTP
            response (for example, ``[('Content-Type', 'application/json')]``). type ``list``
        :param kwargs: Additional arguments (optional). type ``dict``
        :returns: A dict (see :class:`HttpBase` for its structure).
        """
        if headers is None: headers = []

        if 'body' in kwargs:
            if len([x for x in headers if x[0].lower() == "content-type"]) == 0:
                headers.append(("Content-Type", "application/x-www-form-urlencoded"))

            body = kwargs.pop('body')
            if len(kwargs) > 0:
                url = url + UrlEncoded('?' + _encode(**kwargs), skip_encode=True)
        else:
            body = _encode(**kwargs).encode('utf-8')
        message = {
            'method': "POST",
            'headers': headers,
            'body': body
        }
        return self.request(url, message)
    
    def put(self, url, headers=None, **kwargs):
        """
        Send a PUT request to a URL.

        :param url: The URL. type ``string``
        :param headers: A list of pairs specifying the headers for the HTTP
            response (for example, ``[('Content-Type', 'application/json')]``). type ``list``
        :param kwargs: Additional arguments (optional). type ``dict``
        :returns: A dict (see :class:`HtHttpBasetpLib` for its structure).
        """
        if headers is None: headers = []

        if 'body' in kwargs:
            if len([x for x in headers if x[0].lower() == "content-type"]) == 0:
                headers.append(("Content-Type", "application/x-www-form-urlencoded"))

            body = kwargs.pop('body')
            if len(kwargs) > 0:
                url = url + UrlEncoded('?' + _encode(**kwargs), skip_encode=True)
        else:
            body = _encode(**kwargs).encode('utf-8')
        message = {
            'method': "PUT",
            'headers': headers,
            'body': body
        }
        return self.request(url, message)

    def request(self, url, message, **kwargs):
        """
        An HTTP request to a URL.

        :param url: The URL. type ``string``
        :param message: A dict described in :class:`HttpBase`. type ``dict``
        :param kwargs: Additional arguments (optional). type ``dict``
        :returns: A dict (see :class:`HttpBase` for its structure).
        """
        response = self.handler(url, message, **kwargs)
        response = record(response)
        if 400 <= response.status:
            raise HTTPError(response)

        return response

def handler(key_file=None, cert_file=None, timeout=None, verify=False):
    """
    Return an instance of the default HTTP request handler.

    :param `key_file`: Private key path (optional). type ``string``
    :param `cert_file`: Certificate chain file (optional). type ``string``
    :param `timeout`: Request time-out period, in seconds (optional). type ``integer`` or "None"
    :param `verify`: Set to False to disable SSL verification on https connections. type ``Boolean``
    """

    def connect(scheme, host, port):
        kwargs = {}
        if timeout is not None: kwargs['timeout'] = timeout
        if scheme == "http":
            return http.client.HTTPConnection(host, port, **kwargs)
        if scheme == "https":
            if key_file is not None: kwargs['key_file'] = key_file
            if cert_file is not None: kwargs['cert_file'] = cert_file

            if not verify:
                kwargs['context'] = ssl._create_unverified_context()
            return http.client.HTTPSConnection(host, port, **kwargs)
        raise ValueError("unsupported scheme: %s" % scheme)

    def request(url, message, **kwargs):
        scheme, host, port, path = _spliturl(url)
        body = message.get("body", "")
        head = {
            "Content-Length": str(len(body)),
            "Host": host,
            "Accept": "*/*",
            "Connection": "Close",
        } # defaults
        for key, value in message["headers"]:
            head[key] = value
        method = message.get("method", "GET")

        connection = connect(scheme, host, port)
        is_keepalive = False
        try:
            connection.request(method, path, body, head)
            if timeout is not None:
                connection.sock.settimeout(timeout)
            response = connection.getresponse()    
            is_keepalive = "keep-alive" in response.getheader("connection", default="close").lower()        
    
            return {
                "status": response.status,
                "reason": response.reason,
                "headers": response.getheaders(),
                "body": response.read().decode()
            }            
        finally:
            if not is_keepalive:
                connection.close()

    return request
